package at.compax.tools.sql;

import java.io.IOException;
import java.sql.CallableStatement;
import java.sql.Clob;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Statement;
import java.sql.Types;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.List;
import java.util.Vector;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import at.compax.tools.sql.model.ExecutionParameter;
import at.compax.tools.sql.model.QueryParameter;
import at.compax.tools.sql.model.QueryResult;
import at.compax.tools.sql.model.QueryResult.ColumnDefinition;
import at.compax.tools.sql.model.QueryResult.QueryResultBuilder;
import at.compax.tools.sql.model.QueryResult.Row;
import at.compax.tools.sql.model.QueryResult.Row.Column;
import at.compax.tools.sql.model.QueryResult.Row.Column.ColumnBuilder;
import at.compax.tools.sql.model.QueryResult.Row.RowBuilder;
import at.compax.tools.sql.model.Result;
import at.compax.tools.sql.model.Result.ResultBuilder;
import at.compax.tools.sql.model.UserError;
import at.compax.tools.sql.model.UserError.MessageLine;
import at.compax.tools.sql.model.UserError.UserErrorBuilder;
import lombok.Cleanup;
import lombok.extern.log4j.Log4j;

@Log4j
public class SqlExecutor {
  private static final long MAX_ROWS = 10L;

  public static Result executeJson(FrogConnection conn, String json) throws SQLException, JsonParseException, JsonMappingException, IOException {
    log.debug("Parsing JSON: " + json);

    ObjectMapper mapper = new ObjectMapper();
    ExecutionParameter executionParameter = mapper.readValue(json, ExecutionParameter.class);

    boolean isDbOutputOverridden = executionParameter.isEnableDbOutput() && !conn.getDbOutputEnabled();
    if (isDbOutputOverridden) {
      enableDbOutput(conn);
    }

    Result result = execute(conn, executionParameter.getSql());

    if (isDbOutputOverridden) {
      disableDbOutput(conn);
    }

    return result;
  }

  public static Result execute(FrogConnection conn, String sql) throws SQLException, JsonProcessingException {
    log.trace("SQL: " + sql.trim());

    @Cleanup
    Statement stmt = conn.createStatement();

    try {
      String cleanedSql = sql.trim();
      cleanedSql = cleanedSql.replaceAll("/$", "");
      stmt.execute(cleanedSql);
      log.info("Executed");
    } catch (SQLException e) {
      return Result.builder() //
          .exception(e.getClass().getName()) //
          .exceptionMessage(e.getMessage()) //
          .build();
    }

    ResultBuilder resultBuilder = Result.builder();

    try {
      handleUserErrors(conn, sql, stmt);
    } catch (SqlExecutionException e) {
      log.error(e.getMessage(), e);
      resultBuilder.userErrors(e.getUserErrors());
    }

    resultBuilder.dbOutput(parseDbOutput(conn));

    return resultBuilder.build();
  }

  public static QueryResult executeJsonQuery(FrogConnection conn, Long maxRows, String json) throws SQLException, JsonParseException, JsonMappingException, IOException {
    log.debug("Parsing JSON: " + json);

    ObjectMapper mapper = new ObjectMapper();
    QueryParameter queryParameter = mapper.readValue(json, QueryParameter.class);
    log.debug("Parsed QueryParamter: " + queryParameter.toString());

    return executeQuery(conn, maxRows, queryParameter);
  }

  public static QueryResult executeQuery(FrogConnection conn, Long maxRows, QueryParameter parameters) throws SQLException {
    if (parameters == null || parameters.getParameters() == null) {
      return executeQuery(conn, parameters.getSql(), maxRows);
    } else {
      return executeQuery(conn, parameters.getSql(), maxRows, parameters.getParameters().toArray());
    }
  }

  public static QueryResult executeQuery(FrogConnection conn, String sql, Long maxRows, Object... parameters) throws SQLException {
    log.trace("SQL: " + sql.trim());
    if (maxRows == null) {
      maxRows = MAX_ROWS;
    }
    String selectQuery = String.format("SELECT * FROM (%s) WHERE rownum <= %d", sql.replaceAll(";$", "").trim(), maxRows);

    @Cleanup
    PreparedStatement stmt = conn.prepareStatement(selectQuery);

    if (parameters != null) {
      int parameterIndex = 1;
      for (Object object : parameters) {
        if (object != null) {
          if (object instanceof String) {
            stmt.setString(parameterIndex, (String) object);
          } else if (object instanceof Integer) {
            stmt.setInt(parameterIndex, (Integer) object);
          } else if (object instanceof Long) {
            stmt.setLong(parameterIndex, (Long) object);
          } else {
            stmt.setString(parameterIndex, (String) object);
            String message = String.format("Unhandled paramter type <%s>.  Defaulting to String", object.getClass().getName());
            log.warn(message);
          }
        } else {
          log.warn("parameter is null");
        }
        parameterIndex++;
      }
    }

    @Cleanup
    ResultSet rs = null;
    QueryResultBuilder queryResultBuilder = QueryResult.builder();
    try {
      rs = stmt.executeQuery();
    } catch (SQLException e) {
      return queryResultBuilder //
          .exception(e.getClass().getName()) //
          .exceptionMessage(e.getMessage()) //
          .build();
    }

    try {
      handleUserErrors(conn, sql, stmt);
    } catch (SqlExecutionException e) {
      log.error(e.getMessage(), e);

      return queryResultBuilder.userErrors(e.getUserErrors()).build();
    }

    List<Row> rows = new Vector<Row>();
    while (rs.next()) {
      ResultSetMetaData metaData = rs.getMetaData();

      if (rs.isFirst()) {
        queryResultBuilder.columnDefinitions(getColumnDefinitions(metaData));
      }

      RowBuilder rowBuilder = Row.builder();
      List<Column> columns = new Vector<Column>();

      int columnCount = metaData.getColumnCount();
      for (int columnIndex = 1; columnIndex <= columnCount; columnIndex++) {
        ColumnBuilder columnBuilder = Column.builder();

        int columnType = metaData.getColumnType(columnIndex);
        if (rs.getString(columnIndex) == null) {
          columnBuilder.value(null);
        } else if (columnType == Types.VARCHAR) {
          columnBuilder.value(rs.getString(columnIndex));
        } else if (columnType == Types.NUMERIC) {
          columnBuilder.value(rs.getLong(columnIndex));
        } else if (columnType == Types.DATE) {
          columnBuilder.value(rs.getDate(columnIndex));
        } else if (columnType == Types.TIME) {
          columnBuilder.value(rs.getTime(columnIndex));
        } else if (columnType == Types.TIMESTAMP) {
          DateFormat df = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
          columnBuilder.value(df.format(rs.getTimestamp(columnIndex)));
        } else if (columnType == Types.CLOB) {
          Clob clob = rs.getClob(columnIndex);
          columnBuilder.value(clob.getSubString(1, 100));
        } else if (columnType == Types.BLOB) {
          columnBuilder.value("[BLOB]");
        } else {
          String message = String.format("Unhandled column type <%s>.  Defaulting to String", rs.getMetaData().getColumnTypeName(columnIndex));
          log.warn(message);
          columnBuilder.value(rs.getString(columnIndex));
        }

        columns.add(columnBuilder.build());
      }

      rowBuilder.columns(columns);
      rows.add(rowBuilder.build());
    }

    return queryResultBuilder.rows(rows).build();
  }

  public static QueryResult executeUpdate(FrogConnection conn, String sql, Object... parameters) throws SQLException {
    log.trace("SQL: " + sql.trim());
    String updateQuery = sql.replaceAll(";$", "").trim();

    @Cleanup
    PreparedStatement stmt = conn.prepareStatement(updateQuery);

    if (parameters != null) {
      int parameterIndex = 1;
      for (Object object : parameters) {
        if (object != null) {
          if (object instanceof String) {
            stmt.setString(parameterIndex, (String) object);
          } else if (object instanceof Integer) {
            stmt.setInt(parameterIndex, (Integer) object);
          } else if (object instanceof Long) {
            stmt.setLong(parameterIndex, (Long) object);
          } else {
            stmt.setString(parameterIndex, (String) object);
            String message = String.format("Unhandled paramter type <%s>.  Defaulting to String", object.getClass().getName());
            log.warn(message);
          }
        } else {
          log.warn("parameter is null");
        }
        parameterIndex++;
      }
    }

    long affectedRows = 0L;
    QueryResultBuilder queryResultBuilder = QueryResult.builder();
    try {
      affectedRows = stmt.executeUpdate();
    } catch (SQLException e) {
      return queryResultBuilder //
          .exception(e.getClass().getName()) //
          .exceptionMessage(e.getMessage()) //
          .build();
    }

    try {
      handleUserErrors(conn, sql, stmt);
    } catch (SqlExecutionException e) {
      log.error(e.getMessage(), e);

      return queryResultBuilder.userErrors(e.getUserErrors()).build();
    }

    return queryResultBuilder.affectedRows(affectedRows).build();
  }

  private static Vector<ColumnDefinition> getColumnDefinitions(ResultSetMetaData metaData) throws SQLException {
    Vector<ColumnDefinition> columnDefinitions = new Vector<ColumnDefinition>();
    int columnCount = metaData.getColumnCount();

    for (int columnIndex = 1; columnIndex <= columnCount; columnIndex++) {
      columnDefinitions.add(ColumnDefinition.builder() //
          .index(columnIndex) //
          .name(metaData.getColumnLabel(columnIndex)) //
          .type(metaData.getColumnTypeName(columnIndex)) //
          .build());
    }

    return columnDefinitions;
  }

  private static void handleUserErrors(FrogConnection conn, String sql, Statement stmt) throws SQLException, SqlExecutionException {
    SQLWarning warnings = stmt.getWarnings();
    List<UserError> userErrors = new Vector<UserError>();
    if (warnings != null) {
      userErrors = parseUserErrors(conn);
      String message = String.format("Errors while executing SQL \"%s\"", sql);
      throw new SqlExecutionException(message, userErrors);
    }
  }

  private static List<UserError> parseUserErrors(FrogConnection conn) throws SQLException {
    log.debug("Retrieving user errors");
    String userErrorsSql = "select name, type, sequence, line, position, text, attribute, message_number from user_errors where type <> 'JAVA CLASS' order by sequence";
    boolean first = true;

    Vector<UserError> userErrors = new Vector<UserError>();
    Vector<MessageLine> lines = new Vector<UserError.MessageLine>();

    UserErrorBuilder builder = UserError.builder();

    @Cleanup
    PreparedStatement userErrorsStmt = conn.prepareStatement(userErrorsSql);
    @Cleanup
    ResultSet userErrorsRs = userErrorsStmt.executeQuery();

    while (userErrorsRs.next()) {
      long messageNumber = userErrorsRs.getLong("message_number");
      if (messageNumber != 0) {
        if (!first) {
          builder.lines(lines);
          userErrors.add(builder.build());
          first = false;

          builder = UserError.builder();
          lines = new Vector<UserError.MessageLine>();
        }

        builder.name(userErrorsRs.getString("name"));
        builder.type(userErrorsRs.getString("type"));
      }

      lines.add(UserError.MessageLine.builder() //
          .sequence(userErrorsRs.getLong("sequence")) //
          .line(userErrorsRs.getLong("line")) //
          .position(userErrorsRs.getLong("position")) //
          .text(userErrorsRs.getString("text")) //
          .attribute(userErrorsRs.getString("attribute")) //
          .messageNumber(userErrorsRs.getLong("message_number")) //
          .build());
    }

    builder.lines(lines);
    userErrors.add(builder.build());

    return userErrors;
  }

  private static List<String> parseDbOutput(FrogConnection conn) throws SQLException {
    List<String> result = new Vector<>();
    if (conn.getDbOutputEnabled() == null || !conn.getDbOutputEnabled()) {
      return result;
    }

    @Cleanup
    CallableStatement call = conn.prepareCall("{call DBMS_OUTPUT.GET_LINE(?, ?)}");
    call.registerOutParameter(1, Types.VARCHAR);
    call.registerOutParameter(2, Types.NUMERIC);

    long status = 0L;
    while (status == 0L) {
      call.execute();
      status = call.getLong(2);

      String line = call.getString(1);
      log.trace("DBMS_OUTPUT: " + line);
      log.trace("DBMS_OUTPUT Status: " + status);
      if (line != null && status == 0L) {
        result.add(line);
      }
    }
    ;

    return result;

  }

  public static void enableDbOutput(FrogConnection conn) throws SQLException {
    conn.createStatement() //
        .execute("BEGIN DBMS_OUTPUT.ENABLE; END;");

    conn.setDbOutputEnabled(true);
  }

  public static void disableDbOutput(FrogConnection conn) throws SQLException {
    conn.createStatement() //
        .execute("BEGIN DBMS_OUTPUT.DISABLE; END;");

    conn.setDbOutputEnabled(false);
  }
}
