package at.compax.tools.sql.main;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import at.compax.tools.sql.ConnectionParameters;
import at.compax.tools.sql.FrogConnection;
import at.compax.tools.sql.SqlExecutionException;
import at.compax.tools.sql.SqlExecutor;
import at.compax.tools.sql.model.QueryParameter;
import at.compax.tools.sql.model.QueryResult;
import at.compax.tools.sql.model.Result;
import at.compax.tools.sql.model.UserError;
import lombok.Cleanup;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j;

@Log4j
public class Main {
  @Getter
  @RequiredArgsConstructor
  public static enum ExecutionMode {
    PLUGIN_BACKEND(Level.OFF),
    TERMINAL(Level.ALL);

    private final Level loggingLevel;
  }

  private static ExecutionMode executionMode = ExecutionMode.PLUGIN_BACKEND;

  private static FrogConnection getConnection(ConnectionParameters connectionParameters) throws SQLException {
    log.debug("Establishing conneciton...");

    String jdbcConnection = String.format("jdbc:oracle:thin:@%s:%d:%s", connectionParameters.getDbIp(), connectionParameters.getPort(), connectionParameters.getSid());

    Connection sqlConn = DriverManager.getConnection(jdbcConnection, //
        connectionParameters.getUsername(), //
        connectionParameters.getPassword());

    sqlConn.setAutoCommit(false);

    FrogConnection conn = FrogConnection.builder() //
        .conn(sqlConn) //
        .build();
    log.debug("Established conneciton!");

    if (connectionParameters.isEnableDbOutput()) {
      log.info("Enabling db output...");
      try {
        SqlExecutor.enableDbOutput(conn);
      } catch (SQLException e) {
        log.error("Could not enable db output: " + e.getMessage(), e);
        conn.setDbOutputEnabled(false);
      }
    } else {
      conn.setDbOutputEnabled(false);
    }

    return conn;
  }

  private static void closeConnection(FrogConnection conn) {
    if (conn != null) {
      try {
        conn.rollback();
      } catch (SQLException e) {}
      try {
        conn.close();
      } catch (SQLException e) {
        log.error(e);
      }
    }
  }

  private static QueryResult executeExampleQueryWithException(FrogConnection conn) throws SQLException, JsonProcessingException {
    String sql = "SELECT id, version, workspac FROM s_settings";
    return SqlExecutor.executeQuery(conn, sql, 10L);
  }

  private static QueryResult executeExampleQueryWithoutParameter(FrogConnection conn) throws SQLException, JsonProcessingException {
    String sql = "SELECT id, version, workspace FROM s_settings";
    return SqlExecutor.executeQuery(conn, sql, 10L);
  }

  private static QueryResult executeExampleQueryWithParameter(FrogConnection conn) throws SQLException, JsonProcessingException {
    String sql = "select client from k_clients where id = ?";
    return SqlExecutor.executeQuery(conn, sql, 1L);
  }

  private static Result compileExampleSpec(FrogConnection conn) throws SQLException, JsonProcessingException {
    String specSql = "CREATE OR REPLACE PACKAGE aax2_wedel\n" + "AS\n" + "   PROCEDURE create_cli_migration_workflow(p_customer d_customers.id%type,  p_migration_date date);\n" + "END aax2_wedel;";

    return SqlExecutor.execute(conn, specSql);
  }

  private static Result compileExampleSpecWithErrors(FrogConnection conn) throws SQLException, JsonProcessingException {
    String specSql = "CREATE OR REPLACE PACKAGE aax2_wedel\n" + "AS\n" + "   PROCEDURE create_cli_migration_workflow(p_customer d_customers.id%typ,  p_migration_date date);\n" + "END aax2_wedel;";

    return SqlExecutor.execute(conn, specSql);
  }

  private static void handleSqlExecutionException(SqlExecutionException e) throws JsonProcessingException {
    for (UserError userError : e.getUserErrors()) {
      log.debug(userError.toString());
      ObjectMapper mapper = new ObjectMapper();
      System.out.println(mapper.writeValueAsString(userError));
    }
  }

  private static String readFromStdIn() throws IOException {
    BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
    String line;
    StringBuilder sb = new StringBuilder();
    while ((line = br.readLine()) != null) {
      sb.append(line);
    }
    return sb.toString();
  }

  public static void runTests() {
    try {
      // String stdInSql = readFromStdIn();

      @Cleanup
      FrogConnection conn = getConnection(ConnectionParameters.builder() //
          .dbIp("172.19.13.66") //
          .port(1521L) //
          .sid("aax2qc") //
          .username("aax2qc") //
          .password("aax2qc") //
          .build());

      System.out.println(formatJsonObject(SqlExecutor.executeJsonQuery(conn, 10L, "{ \"sql\": \"select customer_id from d_customers\"}")));
      System.out.println(formatJsonObject(compileExampleSpec(conn)));
      System.out.println(formatJsonObject(compileExampleSpecWithErrors(conn)));
      System.out.println(formatJsonObject(executeExampleQueryWithException(conn)));
      System.out.println(formatJsonObject(executeExampleQueryWithoutParameter(conn)));
      System.out.println(formatJsonObject(executeExampleQueryWithParameter(conn)));

      System.out.println(formatJsonObject(SqlExecutor.executeJson(conn, "{ \"sql\": \"begin dbms_output.put_line('should not be seen'); end;\"}")));
      System.out.println(formatJsonObject(SqlExecutor.executeJson(conn, "{ \"enableDbOutput\": true, \"sql\": \"begin dbms_output.put_line('should be seen'); end;\"}")));
      System.out.println(formatJsonObject(SqlExecutor.executeJson(conn, "{ \"sql\": \"begin dbms_output.put_line('should not be seen'); end;\"}")));
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public static String formatJsonObject(Object object) throws JsonProcessingException {
    ObjectMapper mapper = new ObjectMapper();
    if (executionMode == ExecutionMode.TERMINAL) {
      return mapper.writerWithDefaultPrettyPrinter().writeValueAsString(object);
    } else {
      return mapper.writeValueAsString(object);
    }
  }

  private static void serveForever(ConnectionParameters connectionParameters) {
    FrogConnection conn = null;
    try {
      try {
        conn = getConnection(connectionParameters);
      } catch (SQLException e) {
        log.error(e.getMessage(), e);
        System.err.println("Could not connect to Database: " + e.getMessage());
        System.exit(1);
      }

      log.info("Waiting for input...");

      BufferedReader br = new BufferedReader(new InputStreamReader(System.in, Charset.forName("UTF-8")));
      String line;
      StringBuffer commandBuffer = new StringBuffer();

      try {
        while ((line = br.readLine()) != null) {
          if (!"--- END ---".equals(line)) {
            commandBuffer.append(line).append("\n");
          } else {
            ObjectMapper mapper = new ObjectMapper();
            QueryParameter queryParameter = mapper.readValue(commandBuffer.toString(), QueryParameter.class);
            
            String query = queryParameter.getSql();
            String result;
            if (query.toLowerCase().startsWith("select")) {
              result = formatJsonObject(SqlExecutor.executeQuery(conn, query, queryParameter.getRowLimit()));
            } else if (query.toLowerCase().startsWith("update") || query.toLowerCase().startsWith("insert")) {
              result = formatJsonObject(SqlExecutor.executeUpdate(conn, query));
            } else {
              result = formatJsonObject(SqlExecutor.execute(conn, query));
            }
            System.out.println(result);
            commandBuffer.setLength(0);
          }
        }
      } catch (IOException e) {
        log.error(e.getMessage(), e);
      } catch (SQLException e) {
        log.error(e.getMessage(), e);
      }
    } finally {
      closeConnection(conn);
    }
  }

  public static void main(String[] args) {
    String frogRunnerMode = System.getenv("FROG_RUNNER_MODE");
    if (frogRunnerMode != null) {
      if (frogRunnerMode.equals("TERMINAL")) {
        executionMode = ExecutionMode.TERMINAL;
      }
    }

    BasicConfigurator.configure();
    LogManager.getRootLogger().setLevel(Level.OFF);
    LogManager.getLogger("at.compax.tools.sql.main").setLevel(executionMode.getLoggingLevel());
    LogManager.getLogger("at.compax.tools.sql").setLevel(executionMode.getLoggingLevel());

    if (args.length == 0) {
      runTests();
    } else {
      if (args.length < 4)
        throw new IllegalArgumentException("Arguments: Host Port sid username password [max rownums]");

      String host = args[0];
      long port = Long.parseLong(args[1]);
      String sid = args[2];
      String username = args[3];
      String password = args[4];
      serveForever(ConnectionParameters.builder() //
          .dbIp(host) //
          .port(port) //
          .sid(sid) //
          .username(username) //
          .password(password) //
          .build());

    }
  }

}
